"""
Exceptions based on fastapi.exceptions.HTTPException and RFC 6750
"""

from fastapi import HTTPException, status


class InvalidRequestException(HTTPException):
    """Exception raised when request is malformed or invalid."""

    def __init__(self, detail: str) -> None:
        super().__init__(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail={"error": "invalid_request", "message": detail},
        )


class UnauthorizedException(HTTPException):
    """Exception raised when authentication fails (no valid credentials)."""

    def __init__(self, detail: str) -> None:
        super().__init__(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail={"error": "invalid_token", "message": detail},
            headers={"WWW-Authenticate": "Bearer"},
        )


class ForbiddenException(HTTPException):
    """Exception raised when user lacks required permissions."""

    def __init__(self, detail: str) -> None:
        super().__init__(
            status_code=status.HTTP_403_FORBIDDEN,
            detail={"error": "insufficient_scope", "message": detail},
            headers={"WWW-Authenticate": "Bearer"},
        )



"""
Test utilities
"""

import os
from datetime import datetime, timedelta

import jwt
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa

# Global test settings
ZITADEL_ISSUER = os.environ["ZITADEL_HOST"]  # The Zitadel issuer URL
ZITADEL_PROJECT_ID = os.environ["ZITADEL_PROJECT_ID"]  # The project ID where the API app is located
ZITADEL_CLIENT_ID = os.environ["OAUTH_CLIENT_ID"]  # The client ID of the API app
ZITADEL_PRIMARY_DOMAIN = "client-fza.region.zitadel.cloud"  # The primary domain of a Zitadel client


def openid_config_url() -> str:
    """OpenID configuration URL fixture"""
    return f"{ZITADEL_ISSUER}/.well-known/openid-configuration"


def keys_url() -> str:
    """OpenID keys URL fixture"""
    return f"{ZITADEL_ISSUER}/oauth/v2/keys"


def openid_configuration() -> dict:
    """OpenID configuration fixture"""
    return {
        "issuer": ZITADEL_ISSUER,
        "authorization_endpoint": f"{ZITADEL_ISSUER}/oauth/v2/authorize",
        "token_endpoint": f"{ZITADEL_ISSUER}/oauth/v2/token",
        "introspection_endpoint": f"{ZITADEL_ISSUER}/oauth/v2/introspect",
        "userinfo_endpoint": f"{ZITADEL_ISSUER}/oidc/v1/userinfo",
        "revocation_endpoint": f"{ZITADEL_ISSUER}/oauth/v2/revoke",
        "end_session_endpoint": f"{ZITADEL_ISSUER}/oidc/v1/end_session",
        "device_authorization_endpoint": f"{ZITADEL_ISSUER}/oauth/v2/device_authorization",
        "jwks_uri": f"{ZITADEL_ISSUER}/oauth/v2/keys",
        "scopes_supported": [
            "openid",
            "profile",
            "email",
            "phone",
            "address",
            "offline_access",
        ],
        "response_types_supported": ["code", "id_token", "id_token token"],
        "response_modes_supported": ["query", "fragment", "form_post"],
        "grant_types_supported": [
            "authorization_code",
            "implicit",
            "refresh_token",
            "client_credentials",
            "urn:ietf:params:oauth:grant-type:jwt-bearer",
            "urn:ietf:params:oauth:grant-type:device_code",
        ],
        "subject_types_supported": ["public"],
        "id_token_signing_alg_values_supported": ["RS256"],
        "request_object_signing_alg_values_supported": ["RS256"],
        "token_endpoint_auth_methods_supported": [
            "none",
            "client_secret_basic",
            "client_secret_post",
            "private_key_jwt",
        ],
        "token_endpoint_auth_signing_alg_values_supported": ["RS256"],
        "revocation_endpoint_auth_methods_supported": [
            "none",
            "client_secret_basic",
            "client_secret_post",
            "private_key_jwt",
        ],
        "revocation_endpoint_auth_signing_alg_values_supported": ["RS256"],
        "introspection_endpoint_auth_methods_supported": [
            "client_secret_basic",
            "private_key_jwt",
        ],
        "introspection_endpoint_auth_signing_alg_values_supported": ["RS256"],
        "claims_supported": [
            "sub",
            "aud",
            "exp",
            "iat",
            "iss",
            "auth_time",
            "nonce",
            "acr",
            "amr",
            "c_hash",
            "at_hash",
            "act",
            "scopes",
            "client_id",
            "azp",
            "preferred_username",
            "name",
            "family_name",
            "given_name",
            "locale",
            "email",
            "email_verified",
            "phone_number",
            "phone_number_verified",
        ],
        "code_challenge_methods_supported": ["S256"],
        "ui_locales_supported": [
            "bg",
            "cs",
            "de",
            "en",
            "es",
            "fr",
            "hu",
            "id",
            "it",
            "ja",
            "ko",
            "mk",
            "nl",
            "pl",
            "pt",
            "ru",
            "sv",
            "zh",
        ],
        "request_parameter_supported": True,
        "request_uri_parameter_supported": False,
    }


def create_test_token(
    kid: str = "test-key-1",
    expired: bool = False,
    invalid_iss: bool = False,
    invalid_aud: bool = False,
    scopes: str = "scope1",
    evil: bool = False,
    role: str | None = None,
    typ: str = "JWT",
    alg: str = "RS256",
) -> str:
    """Create JWT tokens for testing"""
    now = datetime.now()
    claims = {
        "aud": ["wrong-id"] if invalid_aud else [ZITADEL_PROJECT_ID, ZITADEL_CLIENT_ID],
        "client_id": ZITADEL_CLIENT_ID,
        "exp": int((now - timedelta(hours=1)).timestamp()) if expired else int((now + timedelta(hours=1)).timestamp()),
        "iat": int(now.timestamp()),
        "iss": "wrong-issuer" if invalid_iss else ZITADEL_ISSUER,
        "sub": "user123",
        "nbf": int(now.timestamp()),
        "jti": "unique-token-id",
        "scope": scopes,
    }

    if role:
        claims[f"urn:zitadel:iam:org:project:{ZITADEL_PROJECT_ID}:roles"] = {role: {"role_id": ZITADEL_PRIMARY_DOMAIN}}

    # For evil token use the evil key but claim it's from the valid key
    signing_key = evil_key if evil else valid_key
    headers = {"kid": kid, "typ": typ, "alg": alg}

    private_key = signing_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )

    return jwt.encode(claims, private_key, algorithm="RS256", headers=headers)


def create_openid_keys(empty_keys: bool = False, no_valid_keys: bool = False, additional_key: str = "") -> dict:
    """
    Create OpenID keys
    """
    if empty_keys:
        return {"keys": []}
    elif no_valid_keys:
        # Return a random key that is not valid for the test
        return {
            "keys": [
                {
                    "use": "sig",
                    "kty": "RSA",
                    "kid": "305924551714316751",
                    "alg": "RS256",
                    "n": "rXjVHSfeFS5rtqtDSpdBwolJcteLCDOPVZV8WR0IGrYM7x0fssPimzdWr4PzNY0JvX8CCkYpD99iPhNpwzArC27T2EXrVJmwE93SeyAbwLAXE21h3GIE18UM82y7p7GM-kTc9E9Icvr8UeF9mprARXKVDNq6KdDrwOU70BmUO_FOMBRKpjwIyI1OLIOP69qQ7c6sDLiaQsBHcwolGvMMunzyCLtgWEb6rjRG5wDwu1syqVK4ADWbipFoqx4NGOXzU0yeaiSqnBeu2eJh7r6MTp41IdVz9FPnA0HTXLB3pJZspbDB27g9u1F8RhpDNFYbgcX4YJB6CO6DCKOYEmTxUw",
                    "e": "AQAB",
                }
            ]
        }
    elif additional_key:
        return {
            "keys": [
                {
                    "use": "sig",
                    "kid": "test-key-1",
                    "kty": "RSA",
                    "alg": "RS256",
                    **jwt.algorithms.RSAAlgorithm.to_jwk(
                        valid_key.public_key(),
                        as_dict=True,
                    ),
                },
                {
                    "use": "sig",
                    "kid": additional_key,
                    "kty": "RSA",
                    "alg": "RS256",
                    **jwt.algorithms.RSAAlgorithm.to_jwk(
                        valid_key.public_key(),
                        as_dict=True,
                    ),
                },
            ]
        }
    else:
        return {
            "keys": [
                {
                    "use": "sig",
                    "kid": "test-key-1",
                    "kty": "RSA",
                    "alg": "RS256",
                    **jwt.algorithms.RSAAlgorithm.to_jwk(
                        valid_key.public_key(),
                        as_dict=True,
                    ),
                }
            ]
        }


valid_key = rsa.generate_private_key(backend=default_backend(), public_exponent=65537, key_size=2048)
evil_key = rsa.generate_private_key(backend=default_backend(), public_exponent=65537, key_size=2048)



"""
Test suite for user module.
"""

import time

import pytest
from pydantic import ValidationError

from fastapi_zitadel_auth.user import (
    JwtClaims,
    DefaultZitadelClaims,
    DefaultZitadelUser,
)
from tests.utils import ZITADEL_ISSUER, ZITADEL_PRIMARY_DOMAIN, ZITADEL_CLIENT_ID, ZITADEL_PROJECT_ID

role_key = "role1"
role_id = "295621089671959405"
sub = "22222222222222222222"


@pytest.fixture
def valid_claims_data() -> dict:
    """Fixture providing valid JWT claims data."""
    now = int(time.time())
    return {
        "aud": [ZITADEL_PROJECT_ID],
        "client_id": ZITADEL_CLIENT_ID,
        "exp": now + 3600,
        "iat": now,
        "iss": ZITADEL_ISSUER,
        "sub": sub,
        "nbf": now,
        "jti": "unique-token-id",
    }


@pytest.fixture
def valid_claims_with_project_roles(valid_claims_data):
    """Fixture providing claims data with Zitadel project roles."""
    data = valid_claims_data.copy()
    data[f"urn:zitadel:iam:org:project:{ZITADEL_PROJECT_ID}:roles"] = {role_key: {role_id: ZITADEL_PRIMARY_DOMAIN}}
    return data


class TestBaseZitadelClaims:
    """Test suite for JwtClaims model."""

    @pytest.mark.parametrize("aud", [[ZITADEL_PROJECT_ID], ["audience1", "audience2"]])
    def test_valid_audience_formats(self, valid_claims_data, aud):
        """Test that list audience formats are accepted."""
        data = valid_claims_data.copy()
        data["aud"] = aud
        claims = JwtClaims(**data)
        assert claims.aud == aud

    def test_required_fields(self, valid_claims_data):
        """Test that required fields must be present."""
        required_fields = [
            "aud",
            "client_id",
            "exp",
            "iat",
            "iss",
            "sub",
        ]

        for field in required_fields:
            invalid_data = valid_claims_data.copy()
            del invalid_data[field]

            with pytest.raises(
                ValidationError,
                match=f"1 validation error for JwtClaims\n{field}\n  Field required",
            ):
                JwtClaims(**invalid_data)


class TestDefaultZitadelClaims:
    """Test suite for DefaultZitadelClaims model."""

    def test_project_roles_extraction(self, valid_claims_with_project_roles):
        """Test extraction of project roles from Zitadel-specific claim."""
        claims = DefaultZitadelClaims(**valid_claims_with_project_roles)
        assert claims.project_roles == {role_key: {role_id: ZITADEL_PRIMARY_DOMAIN}}

    def test_missing_project_roles(self, valid_claims_data):
        """Test handling of missing project roles."""
        claims = DefaultZitadelClaims(**valid_claims_data)
        assert claims.project_roles == {}

    def test_different_project_roles(self, valid_claims_data):
        """Test extraction of project roles with different role values."""
        data = valid_claims_data.copy()
        data[f"urn:zitadel:iam:org:project:{ZITADEL_PROJECT_ID}:roles"] = {
            "role2": {"123456789": ZITADEL_PRIMARY_DOMAIN}
        }

        claims = DefaultZitadelClaims(**data)
        assert claims.project_roles == {"role2": {"123456789": ZITADEL_PRIMARY_DOMAIN}}


class TestDefaultZitadelUser:
    """Test suite for DefaultZitadelUser model."""

    def test_valid_user_creation(self, valid_claims_with_project_roles):
        """Test creation of valid DefaultZitadelUser instance."""
        claims = DefaultZitadelClaims(**valid_claims_with_project_roles)
        user = DefaultZitadelUser(claims=claims, access_token="test-token")

        assert isinstance(user.claims, DefaultZitadelClaims)
        assert user.access_token == "test-token"



"""
FastAPI dependencies
"""

from fastapi import Depends

from fastapi_zitadel_auth import ZitadelAuth
from fastapi_zitadel_auth.exceptions import ForbiddenException
from fastapi_zitadel_auth.user import DefaultZitadelUser

try:
    from demo_project.settings import get_settings
except ImportError:
    # ImportError handling since it's also used in tests
    from settings import get_settings


settings = get_settings()

zitadel_auth = ZitadelAuth(
    issuer_url=settings.ZITADEL_HOST,
    project_id=settings.ZITADEL_PROJECT_ID,
    app_client_id=settings.OAUTH_CLIENT_ID,
    allowed_scopes={
        "openid": "OpenID Connect",
        "email": "Email",
        "profile": "Profile",
        "urn:zitadel:iam:org:project:id:zitadel:aud": "Audience",
        "urn:zitadel:iam:org:projects:roles": "Projects roles",
    },
    token_leeway=3,
    scheme_name="ZitadelAuth",
    description="OAuth2 Authorization Code Flow with PKCE for Zitadel authentication",
)


async def validate_is_admin_user(
    user: DefaultZitadelUser = Depends(zitadel_auth),
) -> None:
    """Validate that the authenticated user is a user with a specific role"""
    required_role = "admin"
    if required_role not in user.claims.project_roles:
        raise ForbiddenException(f"User does not have role assigned: {required_role}")



import pytest
from pydantic import BaseModel

from fastapi_zitadel_auth import ZitadelAuth
from fastapi_zitadel_auth.user import (
    BaseZitadelUser,
    JwtClaims,
    DefaultZitadelClaims,
    DefaultZitadelUser,
)
from tests.utils import ZITADEL_ISSUER


class CustomClaims(JwtClaims):
    """Custom claims with additional fields"""

    custom_field: str
    role: str


class CustomUser(BaseZitadelUser):
    """Custom user with specific claims type"""

    username: str


class InvalidClaims(BaseModel):
    """Claims class that doesn't inherit from JwtClaims"""

    some_field: str


class InvalidUser(BaseModel):
    """User class that doesn't inherit from BaseZitadelUser"""

    some_field: str


@pytest.fixture
def default_auth():
    """Fixture for default ZitadelAuth instance"""
    return ZitadelAuth(
        issuer_url=ZITADEL_ISSUER,
        project_id="project_id",
        app_client_id="client_id",
        allowed_scopes={
            "openid": "OpenID Connect",
        },
    )


@pytest.fixture
def custom_auth():
    """Fixture for ZitadelAuth with custom models"""
    return ZitadelAuth(
        claims_model=CustomClaims,
        user_model=CustomUser,
        issuer_url=ZITADEL_ISSUER,
        project_id="project_id",
        app_client_id="client_id",
        allowed_scopes={
            "openid": "OpenID Connect",
        },
    )


class TestZitadelAuth:
    def test_default_initialization(self, default_auth):
        """Test initialization with default models"""
        assert default_auth.claims_model == DefaultZitadelClaims
        assert default_auth.user_model == DefaultZitadelUser

    def test_custom_initialization(self, custom_auth):
        """Test initialization with custom models"""
        assert custom_auth.claims_model == CustomClaims
        assert custom_auth.user_model == CustomUser

    def test_invalid_claims_model(self):
        """Test initialization with invalid claims model"""
        with pytest.raises(ValueError, match="claims_model must be a subclass of JwtClaims"):
            ZitadelAuth(
                claims_model=InvalidClaims,  # type: ignore
                issuer_url=ZITADEL_ISSUER,
                project_id="project_id",
                app_client_id="client_id",
                allowed_scopes={
                    "openid": "OpenID Connect",
                },
            )

    def test_invalid_user_model(self):
        """Test initialization with invalid user model"""
        with pytest.raises(ValueError, match="user_model must be a subclass of BaseZitadelUser"):
            ZitadelAuth(
                user_model=InvalidUser,  # type: ignore
                issuer_url=ZITADEL_ISSUER,
                project_id="project_id",
                app_client_id="client_id",
                allowed_scopes={
                    "openid": "OpenID Connect",
                },
            )

    def test_user_with_claims(self):
        """Test creation of user with claims"""
        claims_data = {
            "aud": "test_client",
            "client_id": "123",
            "exp": 1234567890,
            "iat": 1234567890,
            "iss": "test_issuer",
            "sub": "user123",
            "nbf": 1234567890,
            "jti": "token123",
            "custom_field": "custom_value",
            "role": "admin",
        }
        claims = CustomClaims(**claims_data)
        user = CustomUser(claims=claims, access_token="test_token", username="testuser")
        assert user.claims == claims
        assert user.access_token == "test_token"
        assert user.username == "testuser"
        assert user.claims.custom_field == "custom_value"

    def test_empty_scheme_name(self):
        """Test initialization with empty scheme_name raises ValueError"""
        with pytest.raises(ValueError, match="scheme_name must be a non-empty string"):
            ZitadelAuth(
                issuer_url=ZITADEL_ISSUER,
                project_id="project_id",
                app_client_id="client_id",
                allowed_scopes={"openid": "OpenID Connect"},
                scheme_name="",
            )

    def test_empty_description(self):
        """Test initialization with empty description raises ValueError"""
        with pytest.raises(ValueError, match="description must be a non-empty string"):
            ZitadelAuth(
                issuer_url=ZITADEL_ISSUER,
                project_id="project_id",
                app_client_id="client_id",
                allowed_scopes={"openid": "OpenID Connect"},
                description="",
            )

    def test_whitespace_scheme_name(self):
        """Test initialization with whitespace scheme_name raises ValueError"""
        with pytest.raises(ValueError, match="scheme_name must be a non-empty string"):
            ZitadelAuth(
                issuer_url=ZITADEL_ISSUER,
                project_id="project_id",
                app_client_id="client_id",
                allowed_scopes={"openid": "OpenID Connect"},
                scheme_name="   ",
            )

    def test_whitespace_description(self):
        """Test initialization with whitespace description raises ValueError"""
        with pytest.raises(ValueError, match="description must be a non-empty string"):
            ZitadelAuth(
                issuer_url=ZITADEL_ISSUER,
                project_id="project_id",
                app_client_id="client_id",
                allowed_scopes={"openid": "OpenID Connect"},
                description="   ",
            )

    def test_custom_scheme_name_and_description(self):
        """Test initialization with custom scheme_name and description"""
        custom_auth = ZitadelAuth(
            issuer_url=ZITADEL_ISSUER,
            project_id="project_id",
            app_client_id="client_id",
            allowed_scopes={"openid": "OpenID Connect"},
            scheme_name="CustomAuthScheme",
            description="Custom authentication description",
        )
        assert custom_auth.scheme_name == "CustomAuthScheme"



"""
Pytest conftest.py file to define fixtures available to all tests
"""

from typing import Iterator

import httpx
import pytest
from blockbuster import blockbuster_ctx, BlockBuster
from starlette.testclient import TestClient

from demo_project.dependencies import zitadel_auth
from demo_project.main import app
from fastapi_zitadel_auth import ZitadelAuth
from tests.utils import (
    create_openid_keys,
    openid_config_url,
    openid_configuration,
    keys_url,
    ZITADEL_CLIENT_ID,
    ZITADEL_ISSUER,
    ZITADEL_PROJECT_ID,
)


@pytest.fixture
def fastapi_app():
    """FastAPI app fixture"""
    zitadel_auth_overrides = ZitadelAuth(
        issuer_url=ZITADEL_ISSUER,
        app_client_id=ZITADEL_CLIENT_ID,
        project_id=ZITADEL_PROJECT_ID,
        allowed_scopes={"scope1": "Some scope"},
    )
    app.dependency_overrides[zitadel_auth] = zitadel_auth_overrides
    yield


@pytest.fixture(autouse=True)
def blockbuster() -> Iterator[BlockBuster]:
    """Detect blocking calls within an asynchronous event loop"""
    with blockbuster_ctx() as bb:
        yield bb


@pytest.fixture(autouse=True)
async def reset_openid_config():
    """Reset the OpenID configuration before each test"""
    zitadel_auth.openid_config.reset_cache()
    yield


@pytest.fixture
def mock_openid(respx_mock):
    """Fixture to mock OpenID configuration"""
    respx_mock.get(openid_config_url()).respond(json=openid_configuration())
    yield


@pytest.fixture
def mock_openid_and_keys(respx_mock, mock_openid):
    """Fixture to mock OpenID configuration and keys"""
    respx_mock.get(keys_url()).respond(json=create_openid_keys())
    yield


@pytest.fixture
def mock_openid_and_empty_keys(respx_mock, mock_openid):
    """Fixture to mock OpenID configuration and empty keys"""
    respx_mock.get(keys_url()).respond(json=create_openid_keys(empty_keys=True))
    yield


@pytest.fixture
def mock_openid_empty_then_ok(respx_mock, mock_openid):
    keys_route = respx_mock.get(keys_url())
    keys_route.side_effect = [
        httpx.Response(json=create_openid_keys(empty_keys=True), status_code=200),
        httpx.Response(json=create_openid_keys(additional_key="test-key-2"), status_code=200),
    ]
    openid_route = respx_mock.get(openid_config_url())
    openid_route.side_effect = [
        httpx.Response(json=openid_configuration(), status_code=200),
        httpx.Response(json=openid_configuration(), status_code=200),
    ]
    yield
    assert keys_route.call_count == 2
    assert openid_route.call_count == 2


@pytest.fixture
def mock_openid_and_no_valid_keys(respx_mock, mock_openid):
    """Fixture to mock OpenID configuration and keys with no valid keys"""
    respx_mock.get(keys_url()).respond(json=create_openid_keys(no_valid_keys=True))
    yield


@pytest.fixture
def public_client():
    """Test client that does not run startup event."""
    yield TestClient(app=app)



"""
Sample FastAPI app with Zitadel authentication
"""

import logging
from contextlib import asynccontextmanager
from typing import AsyncIterator

import uvicorn
from fastapi import FastAPI, Request, Security
from starlette.middleware.cors import CORSMiddleware

from fastapi_zitadel_auth import __version__


try:
    from demo_project.dependencies import zitadel_auth, validate_is_admin_user  # type: ignore[no-redef]
    from demo_project.settings import get_settings  # type: ignore[no-redef]
except ImportError:
    # ImportError handling since it's also used in tests
    from dependencies import zitadel_auth, validate_is_admin_user  # type: ignore[no-redef]
    from settings import get_settings  # type: ignore[no-redef]

settings = get_settings()

# setup logging
logger = logging.getLogger(__name__)
logging.basicConfig(level=settings.LOG_LEVEL)
logging.getLogger("httpcore").setLevel(logging.WARNING)
logging.getLogger("httpx").setLevel(logging.WARNING)
logging.getLogger("hpack").setLevel(logging.WARNING)
logger.debug(f"Settings: {settings.model_dump_json()}")


@asynccontextmanager
async def lifespan(app: FastAPI) -> AsyncIterator[None]:
    """
    Load OpenID config on startup.
    """
    await zitadel_auth.openid_config.load_config()
    yield


app = FastAPI(
    title="fastapi-zitadel-auth demo",
    lifespan=lifespan,
    version=__version__,
    swagger_ui_oauth2_redirect_url="/oauth2-redirect",
    swagger_ui_init_oauth={
        "usePkceWithAuthorizationCodeGrant": True,
        "clientId": settings.OAUTH_CLIENT_ID,
        "scopes": " ".join(
            [
                "openid",
                "profile",
                "email",
                "urn:zitadel:iam:org:projects:roles",
                "urn:zitadel:iam:org:project:id:zitadel:aud",
            ]
        ),
    },
)

# Middleware
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


@app.get("/api/public", summary="Public endpoint")
def public():
    """Public endpoint"""
    return {"message": "Hello world!"}


@app.get(
    "/api/protected/admin",
    summary="Protected endpoint, requires admin role",
    dependencies=[Security(validate_is_admin_user)],
)
def protected_for_admin(request: Request):
    """Protected endpoint"""
    user = request.state.user
    return {"message": "Hello world!", "user": user}


@app.get(
    "/api/protected/scope",
    summary="Protected endpoint, requires a specific scope",
    dependencies=[Security(zitadel_auth, scopes=["scope1"])],
)
def protected_by_scope(request: Request):
    """Protected endpoint, requires a specific scope"""
    user = request.state.user
    return {"message": "Hello world!", "user": user}


if __name__ == "__main__":
    uvicorn.run("main:app", reload=True, port=8001)



from typing import (
    TypeVar,
    Generic,
    Any,  # noqa
)

from pydantic import BaseModel, Field, model_validator

# Generic type variables for claims and user models
ClaimsT = TypeVar("ClaimsT", bound="JwtClaims")
UserT = TypeVar("UserT", bound="BaseZitadelUser[Any]")


class JwtClaims(BaseModel):
    """Base model for JWT access token claims in Zitadel
    as per RFC 7519 and Zitadel
    """

    aud: str | list[str] = Field(description="The audience of the token (e.g. app_client_id and project_id)")
    client_id: str = Field(description="Client id of the client who requested the token")
    exp: int = Field(description="Time the token expires (as unix time)")
    iat: int = Field(description="Time of the token was issued at (as unix time)")
    iss: str = Field(description="Issuing domain of a token")
    sub: str = Field(description="Subject ID of the user")
    nbf: int | None = Field(default=None, description="Time the token must not be used before (as unix time)")
    jti: str | None = Field(default=None, description="Unique id of the token")


class BaseZitadelUser(BaseModel, Generic[ClaimsT]):
    """Base authenticated user with claims and token"""

    claims: ClaimsT
    access_token: str


class DefaultZitadelClaims(JwtClaims):
    """Default Zitadel claims implementation with project roles"""

    project_roles: dict[str, dict[str, str]] = Field(
        default_factory=dict,
    )

    @model_validator(mode="before")
    @classmethod
    def extract_project_roles(cls, values: dict) -> dict:
        """Extract project-specific role claim into project_roles field"""
        for key in values.keys():
            if key.startswith("urn:zitadel:iam:org:project:") and key.endswith(":roles"):
                values["project_roles"] = values[key]
                break
        return values


class DefaultZitadelUser(BaseZitadelUser[DefaultZitadelClaims]):
    """Default Zitadel user implementation"""

    claims: DefaultZitadelClaims



import asyncio
from asyncio import Lock
from datetime import datetime, timedelta
import logging
from typing import Any
import httpx
from jwt import PyJWK
from pydantic import BaseModel, ConfigDict, PositiveInt
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey

from fastapi_zitadel_auth.exceptions import UnauthorizedException

log = logging.getLogger("fastapi_zitadel_auth")


class OpenIdConfig(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, strict=True, extra="forbid")

    issuer_url: str
    config_url: str
    authorization_url: str
    token_url: str
    jwks_uri: str
    signing_keys: dict[str, RSAPublicKey] = {}

    refresh_lock: Lock = Lock()
    last_refresh_timestamp: datetime | None = None
    cache_ttl_seconds: PositiveInt = 600

    async def load_config(self) -> None:
        """Refresh the openid configuration and signing keys if necessary."""
        async with self.refresh_lock:
            if not self._needs_refresh():
                return
            log.debug("Loading OpenID configuration.")
            current_time = datetime.now()
            try:
                async with httpx.AsyncClient(timeout=10, http2=True) as client:
                    config = await self._fetch_config(client)
                    new_keys = await self._fetch_jwks(client)

                self.issuer_url = config["issuer"]
                self.authorization_url = config["authorization_endpoint"]
                self.token_url = config["token_endpoint"]
                self.jwks_uri = config["jwks_uri"]
                self.signing_keys = self._parse_jwks(new_keys)
                self.last_refresh_timestamp = current_time

            except Exception as e:
                log.exception(f"Unable to refresh configuration from identity provider: {e}")
                self.reset_cache()
                raise UnauthorizedException("Unable to refresh configuration from identity provider")

        log.info("fastapi-zitadel-auth loaded OpenID configuration and signing keys from Zitadel.")
        log.info("Issuer:               %s", self.issuer_url)
        log.info("Authorization url:    %s", self.authorization_url)
        log.info("Token url:            %s", self.token_url)
        log.debug("Keys url:            %s", self.jwks_uri)
        log.debug("Last refresh:        %s", self.last_refresh_timestamp)
        log.debug("Signing keys:        %s", len(self.signing_keys))
        log.debug("Cache TTL:           %s s", self.cache_ttl_seconds)

    async def get_key(self, kid: str) -> RSAPublicKey:
        """Get a signing key by its ID, refreshing JWKS once if necessary."""
        if kid not in self.signing_keys:
            log.debug("Key '%s' not found, refreshing JWKS", kid)
            await self._sleep()
            self.reset_cache()
            await self.load_config()

            if kid not in self.signing_keys:
                log.error(f"Unable to verify token, no signing keys found for key with ID: '{kid}'")
                raise UnauthorizedException("Unable to verify token, no signing keys found")
        return self.signing_keys[kid]

    def reset_cache(self) -> None:
        """Reset the cache by clearing the timestamp and keys."""
        self.last_refresh_timestamp = None
        self.signing_keys = {}
        log.debug("Reset OpenID configuration cache")

    @staticmethod
    async def _sleep() -> None:
        """Wait for a short period to allow other tasks to run."""
        log.debug("Waiting for other tasks to finish...")
        await asyncio.sleep(1)

    def _needs_refresh(self) -> bool:
        """Check if the cached keys should be refreshed based on cache state or time elapsed."""
        if not self.last_refresh_timestamp or not self.signing_keys:
            return True

        elapsed = datetime.now() - self.last_refresh_timestamp
        return elapsed > timedelta(seconds=self.cache_ttl_seconds)

    async def _fetch_config(self, client: httpx.AsyncClient) -> dict[str, Any]:
        """Fetch OpenID Connect configuration."""
        log.info("Fetching OpenID configuration from %s", self.config_url)
        response = await client.get(self.config_url)
        response.raise_for_status()
        return response.json()

    async def _fetch_jwks(self, client: httpx.AsyncClient) -> dict[str, list]:
        """Fetch JWK Set from the jwks_uri."""
        log.info("Fetching JWKS from %s", self.jwks_uri)
        response = await client.get(self.jwks_uri)
        response.raise_for_status()
        return response.json()

    @staticmethod
    def _parse_jwks(jwks: dict[str, list]) -> dict[str, RSAPublicKey]:
        """Parse the JWKS response and return a dictionary of RSA public keys."""
        keys = {}
        available_keys = jwks.get("keys", [])
        for key in available_keys:
            if key.get("use") == "sig" and key.get("alg") == "RS256" and key.get("kty") == "RSA" and "kid" in key:
                log.debug("Loading public key %s", key)
                keys[key["kid"]] = PyJWK.from_dict(obj=key, algorithm="RS256").key
        return keys



"""
Test the TokenValidator class
"""

import time

import jwt
import pytest
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa

from fastapi_zitadel_auth.exceptions import UnauthorizedException, ForbiddenException
from fastapi_zitadel_auth.token import TokenValidator
from tests.utils import ZITADEL_ISSUER, ZITADEL_CLIENT_ID, ZITADEL_PROJECT_ID


@pytest.fixture(scope="module")
def rsa_keys() -> tuple:
    """Generate RSA key pair"""
    private_key = rsa.generate_private_key(backend=default_backend(), public_exponent=65537, key_size=2048)
    public_key = private_key.public_key()
    return private_key, public_key


@pytest.fixture
def token_validator() -> TokenValidator:
    """TokenValidator fixture"""
    return TokenValidator()


@pytest.fixture
def valid_token(rsa_keys) -> str:
    """Generate a valid JWT token"""
    private_key, _ = rsa_keys
    now = int(time.time())

    claims = {
        "sub": "user123",
        "iss": ZITADEL_ISSUER,
        "aud": [ZITADEL_CLIENT_ID, ZITADEL_PROJECT_ID],
        "exp": now + 3600,
        "iat": now,
        "nbf": now,
        "jti": "unique-token-id",
    }

    pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )

    return jwt.encode(claims, pem, algorithm="RS256", headers={"kid": "test-key-1"})


class TestTokenValidator:
    """Test the TokenValidator class"""

    @pytest.mark.parametrize(
        "claims,required_scopes,expected",
        [
            ({"scope": "read:messages write:messages"}, None, True),
            ({"scope": "read:messages write:messages"}, ["read:messages"], True),
            (
                {"scope": "read:messages write:messages"},
                ["read:messages", "write:messages"],
                True,
            ),
            (
                {"scope": "read:messages"},
                ["write:messages"],
                pytest.raises(ForbiddenException),
            ),
            ({"scope": ""}, ["read:messages"], pytest.raises(ForbiddenException)),
            ({}, ["read:messages"], pytest.raises(ForbiddenException)),
            (
                {"scope": "read:messages write:messages"},
                ["read:messages", "delete:messages"],
                pytest.raises(ForbiddenException),
            ),
            # Test with multiple space-separated allowed_scopes
            ({"scope": "scope1 scope2 scope3"}, ["scope2"], True),
            # Test with empty required allowed_scopes list
            ({"scope": "read:messages"}, [], True),
            # Test with special characters in allowed_scopes
            (
                {"scope": "api:read user.profile system-admin"},
                ["api:read", "system-admin"],
                True,
            ),
        ],
    )
    def test_validate_scopes(self, claims, required_scopes, expected):
        """Test scope validation with various combinations of claims and required allowed_scopes"""
        if isinstance(expected, bool):
            assert TokenValidator.validate_scopes(claims, required_scopes) == expected
        else:
            with expected:
                TokenValidator.validate_scopes(claims, required_scopes)

    @pytest.mark.parametrize(
        "claims",
        [
            {"scope": None},
            {"scope": 123},
            {"scope": True},
            {"scope": {"invalid": "type"}},
        ],
    )
    def test_validate_scopes_invalid_type(self, claims):
        """Test scope validation with invalid scope types"""
        with pytest.raises(UnauthorizedException):
            TokenValidator.validate_scopes(claims, ["read:messages"])

    def test_validate_scopes_whitespace_handling(self):
        """Test handling of various whitespace patterns in scope strings"""
        claims = {"scope": "scope1    scope2\tscope3\n\rscope4"}
        assert TokenValidator.validate_scopes(claims, ["scope1", "scope4"]) is True

    @pytest.mark.parametrize(
        "scope_string,required_scope",
        [
            ("a" * 1000 + " valid:scope", "valid:scope"),  # Very long scope string
            ("scope1" + " " * 100 + "scope2", "scope2"),  # Multiple spaces
            ("\u3000scope1\u2000scope2", "scope1"),  # Unicode whitespace
        ],
    )
    def test_validate_scopes_edge_cases(self, scope_string, required_scope):
        """Test scope validation with edge cases"""
        claims = {"scope": scope_string}
        assert TokenValidator.validate_scopes(claims, [required_scope]) is True

    def test_parse_unverified_valid_token(self, token_validator, valid_token):
        """
        Test that the TokenValidator can parse an unverified token
        """
        header, claims = token_validator.parse_unverified_token(valid_token)

        assert isinstance(header, dict)
        assert isinstance(claims, dict)
        assert header["kid"] == "test-key-1"
        assert header["alg"] == "RS256"
        assert claims["sub"] == "user123"
        assert "exp" in claims
        assert "iat" in claims
        assert "iss" in claims
        assert "aud" in claims
        assert "nbf" in claims
        assert "jti" in claims

    def test_parse_unverified_none_token(self, token_validator):
        """Test that the TokenValidator raises an exception when parsing a None token"""
        with pytest.raises(UnauthorizedException, match="Invalid token format"):
            token_validator.parse_unverified_token(None)  # type: ignore

    @pytest.mark.parametrize(
        "invalid_token",
        [
            "not.a.token",
            "invalid.token.format",
            "eyJhbGciOiJIUzI1NiJ9",  # Only header
            "",
            "null",
            None,
        ],
        ids=[
            "random_string",
            "wrong_format",
            "header_only",
            "empty_string",
            "string_null",
            "none_value",
        ],
    )
    def test_parse_unverified_invalid_token(self, token_validator, invalid_token):
        """Test that the TokenValidator raises an exception when parsing an invalid token"""
        with pytest.raises(UnauthorizedException, match="Invalid token format"):
            token_validator.parse_unverified_token(invalid_token)

    def test_verify_valid_token(self, token_validator, valid_token, rsa_keys):
        """Test that the TokenValidator can verify a valid token"""
        _, public_key = rsa_keys

        claims = token_validator.verify(
            token=valid_token,
            key=public_key,
            audiences=[ZITADEL_CLIENT_ID, ZITADEL_PROJECT_ID],
            issuer=ZITADEL_ISSUER,
        )

        assert claims["sub"] == "user123"
        assert claims["iss"] == ZITADEL_ISSUER
        assert ZITADEL_CLIENT_ID in claims["aud"]

    def test_verify_expired_token(self, token_validator, rsa_keys):
        """Test that the TokenValidator raises an exception when verifying an expired token"""
        private_key, public_key = rsa_keys
        now = int(time.time())

        expired_claims = {
            "sub": "user123",
            "iss": ZITADEL_ISSUER,
            "aud": [ZITADEL_CLIENT_ID],
            "exp": now - 3600,  # Expired 1 hour ago
            "iat": now - 7200,
            "nbf": now - 7200,
            "jti": "unique-token-id",
        }

        pem = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        )

        expired_token = jwt.encode(expired_claims, pem, algorithm="RS256")

        with pytest.raises(jwt.ExpiredSignatureError):
            token_validator.verify(
                token=expired_token,
                key=public_key,
                audiences=[ZITADEL_CLIENT_ID],
                issuer=ZITADEL_ISSUER,
            )

    def test_verify_invalid_audience(self, token_validator, valid_token, rsa_keys):
        """Test that the TokenValidator raises an exception when verifying a token with an invalid audience"""
        _, public_key = rsa_keys

        with pytest.raises(jwt.InvalidAudienceError):
            token_validator.verify(
                token=valid_token,
                key=public_key,
                audiences=["wrong_audience"],
                issuer=ZITADEL_ISSUER,
            )

    def test_verify_invalid_issuer(self, token_validator, valid_token, rsa_keys):
        """Test that the TokenValidator raises an exception when verifying a token with an invalid issuer_url"""
        _, public_key = rsa_keys

        with pytest.raises(jwt.InvalidIssuerError):
            token_validator.verify(
                token=valid_token,
                key=public_key,
                audiences=[ZITADEL_CLIENT_ID, ZITADEL_PROJECT_ID],
                issuer="https://wrong.issuer.com",
            )

    def test_verify_invalid_signature(self, token_validator, valid_token):
        """Test that the TokenValidator raises an exception when verifying a token with an invalid signature"""
        wrong_key = rsa.generate_private_key(
            backend=default_backend(), public_exponent=65537, key_size=2048
        ).public_key()

        with pytest.raises(jwt.InvalidSignatureError):
            token_validator.verify(
                token=valid_token,
                key=wrong_key,
                audiences=[ZITADEL_CLIENT_ID, ZITADEL_PROJECT_ID],
                issuer=ZITADEL_ISSUER,
            )

    def test_verify_not_yet_valid(self, token_validator, rsa_keys):
        """Raise Exception when verifying a token that is not yet valid"""
        private_key, public_key = rsa_keys
        now = int(time.time())

        future_claims = {
            "sub": "user123",
            "iss": ZITADEL_ISSUER,
            "aud": [ZITADEL_CLIENT_ID],
            "exp": now + 7200,
            "iat": now,
            "nbf": now + 3600,  # Not valid for another hour
            "jti": "unique-token-id",
        }

        pem = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        )

        future_token = jwt.encode(future_claims, pem, algorithm="RS256")

        with pytest.raises(jwt.ImmatureSignatureError):
            token_validator.verify(
                token=future_token,
                key=public_key,
                audiences=[ZITADEL_CLIENT_ID],
                issuer=ZITADEL_ISSUER,
            )

    def test_verify_missing_claims(self, token_validator, rsa_keys):
        """Raise Exception when verifying a token with missing required claims"""
        private_key, public_key = rsa_keys
        now = int(time.time())

        incomplete_claims = {
            "iss": ZITADEL_ISSUER,
            "aud": [ZITADEL_CLIENT_ID],
            "exp": now + 3600,
            # Missing 'sub' claim
        }

        pem = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        )

        incomplete_token = jwt.encode(incomplete_claims, pem, algorithm="RS256")

        with pytest.raises(jwt.MissingRequiredClaimError):
            token_validator.verify(
                token=incomplete_token,
                key=public_key,
                audiences=[ZITADEL_CLIENT_ID],
                issuer=ZITADEL_ISSUER,
            )



"""
This module demonstrates how to authenticate a service user with Zitadel.
"""

import asyncio
import json
import time
import jwt as pyjwt
from httpx import AsyncClient

try:
    from settings import get_settings
except ImportError:
    # ImportError handling since it's also used in tests
    from demo_project.settings import get_settings

# UPDATE THESE VALUES ----------------------------------------------
# The service account private key file downloaded from Zitadel
SERVICE_USER_PRIVATE_KEY_FILE = "service_user.json"

# The role that the service user should have in Zitadel
ROLE = "admin"
# ---------------------------------------------------------------

# Loading the service account private key JSON file
with open(SERVICE_USER_PRIVATE_KEY_FILE, "r") as file:
    json_data = json.load(file)


# get settings from the settings module for simplicity
settings = get_settings()

# Extracting necessary values from the JSON data
private_key = json_data["key"]
kid = json_data["keyId"]
user_id = json_data["userId"]

# Preparing the JWT header and payload for authentication
header = {"alg": "RS256", "kid": kid}
payload = {
    "iss": user_id,
    "sub": user_id,
    "aud": str(settings.ZITADEL_HOST).rstrip("/"),  # Remove trailing slash
    "iat": int(time.time()),
    "exp": int(time.time()) + 3600,  # Token expires in 1 hour
}

# Generating JWT token
jwt_token = pyjwt.encode(payload, private_key, algorithm="RS256", headers=header)


async def main():
    """
    Authenticate a service user with Zitadel and make an API call to a protected endpoint.
    """
    # Creating an asynchronous HTTP client context
    async with AsyncClient() as client:
        # Data payload for the OAuth2 token request
        data = {
            "grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer",
            "scope": " ".join(
                [
                    "openid",
                    f"urn:zitadel:iam:org:project:role:{ROLE}",
                    f"urn:zitadel:iam:org:project:id:{settings.ZITADEL_PROJECT_ID}:aud",
                ]
            ),
            "assertion": jwt_token,
        }

        # Making a POST request to the OAuth2 token endpoint
        response = await client.post(url=f"{settings.ZITADEL_HOST}oauth/v2/token", data=data)

        # Handling the response
        response.raise_for_status()
        access_token = response.json()["access_token"]

        # Example API call using the acquired access token
        my_api_response = await client.get(
            "http://localhost:8001/api/protected/admin",
            headers={"Authorization": f"Bearer {access_token}"},
        )
        if my_api_response.status_code == 200:
            print(my_api_response.json())
        else:
            print(f"Error: {my_api_response.status_code} - {my_api_response.text}")


if __name__ == "__main__":
    asyncio.run(main())



"""
Test the auth module with endpoint tests.
"""

import logging
import time

import pytest
from httpx import ASGITransport, AsyncClient

from demo_project.main import app
from fastapi_zitadel_auth import ZitadelAuth
from fastapi_zitadel_auth.token import TokenValidator
from tests.utils import (
    create_test_token,
    ZITADEL_PRIMARY_DOMAIN,
    ZITADEL_ISSUER,
    ZITADEL_PROJECT_ID,
    ZITADEL_CLIENT_ID,
)

log = logging.getLogger("fastapi_zitadel_auth")


@pytest.mark.asyncio
async def test_admin_user(fastapi_app, mock_openid_and_keys):
    """Test that with a valid token we can access the protected endpoint."""
    issued_at = int(time.time())
    expires = issued_at + 3600
    access_token = create_test_token(role="admin")
    async with AsyncClient(
        transport=ASGITransport(app=app),
        base_url="http://test",
        headers={"Authorization": f"Bearer {access_token}"},
    ) as ac:
        response = await ac.get("/api/protected/admin")
        assert response.status_code == 200, response.text
        # see create_test_token for the claims
        assert response.json() == {
            "message": "Hello world!",
            "user": {
                "access_token": access_token,
                "claims": {
                    "aud": [
                        ZITADEL_PROJECT_ID,
                        ZITADEL_CLIENT_ID,
                    ],
                    "client_id": ZITADEL_CLIENT_ID,
                    "exp": expires,
                    "iat": issued_at,
                    "iss": ZITADEL_ISSUER,
                    "jti": "unique-token-id",
                    "nbf": issued_at,
                    "project_roles": {
                        "admin": {
                            "role_id": ZITADEL_PRIMARY_DOMAIN,
                        },
                    },
                    "sub": "user123",
                },
            },
        }


async def test_no_keys_to_decode_with(fastapi_app, mock_openid_and_empty_keys):
    """Test that if no signing keys are found, the token cannot be decoded."""
    async with AsyncClient(
        transport=ASGITransport(app=app),
        base_url="http://test",
        headers={"Authorization": "Bearer " + create_test_token(role="admin")},
    ) as ac:
        response = await ac.get("/api/protected/admin")
        assert response.status_code == 401
        assert response.json() == {
            "detail": {"error": "invalid_token", "message": "Unable to verify token, no signing keys found"}
        }
        assert response.headers["WWW-Authenticate"] == "Bearer"


async def test_normal_user_rejected(fastapi_app, mock_openid_and_keys):
    """Test that a user without the admin role is rejected from the admin endpoint."""
    async with AsyncClient(
        transport=ASGITransport(app=app),
        base_url="http://test",
        headers={"Authorization": "Bearer " + create_test_token(role="user")},
    ) as ac:
        response = await ac.get("/api/protected/admin")
        assert response.status_code == 403
        assert response.json() == {
            "detail": {"error": "insufficient_scope", "message": "User does not have role assigned: admin"}
        }
        assert response.headers["WWW-Authenticate"] == "Bearer"


async def test_invalid_token_issuer(fastapi_app, mock_openid_and_keys):
    """Test that a token with an invalid issuer is rejected."""
    async with AsyncClient(
        transport=ASGITransport(app=app),
        base_url="http://test",
        headers={"Authorization": "Bearer " + create_test_token(role="admin", invalid_iss=True)},
    ) as ac:
        response = await ac.get("/api/protected/admin")
        assert response.status_code == 401
        assert response.json() == {"detail": {"error": "invalid_token", "message": "Token contains invalid claims"}}
        assert response.headers["WWW-Authenticate"] == "Bearer"


async def test_invalid_token_audience(fastapi_app, mock_openid_and_keys):
    """Test that a token with an invalid audience is rejected."""
    async with AsyncClient(
        transport=ASGITransport(app=app),
        base_url="http://test",
        headers={"Authorization": "Bearer " + create_test_token(role="admin", invalid_aud=True)},
    ) as ac:
        response = await ac.get("/api/protected/admin")
        assert response.status_code == 401
        assert response.json() == {"detail": {"error": "invalid_token", "message": "Token contains invalid claims"}}
        assert response.headers["WWW-Authenticate"] == "Bearer"


async def test_no_valid_keys_for_token(fastapi_app, mock_openid_and_no_valid_keys):
    """Test that if no valid keys are found, the token cannot be decoded."""
    async with AsyncClient(
        transport=ASGITransport(app=app),
        base_url="http://test",
        headers={"Authorization": "Bearer " + create_test_token(role="admin")},
    ) as ac:
        response = await ac.get("/api/protected/admin")
        assert response.status_code == 401
        assert response.json() == {
            "detail": {"error": "invalid_token", "message": "Unable to verify token, no signing keys found"}
        }
        assert response.headers["WWW-Authenticate"] == "Bearer"


async def test_no_valid_scopes(fastapi_app, mock_openid_and_keys):
    """Test that a token without the required scopes is rejected."""
    async with AsyncClient(
        transport=ASGITransport(app=app),
        base_url="http://test",
        headers={"Authorization": "Bearer " + create_test_token(scopes="openid email profile")},
    ) as ac:
        response = await ac.get("/api/protected/scope")
    assert response.status_code == 403
    assert response.json() == {"detail": {"error": "insufficient_scope", "message": "Missing required scope: scope1"}}
    assert response.headers["WWW-Authenticate"] == "Bearer"


async def test_invalid_scopes_format(fastapi_app, mock_openid_and_keys):
    """Test that a token with invalid scope format is rejected."""
    async with AsyncClient(
        transport=ASGITransport(app=app),
        base_url="http://test",
        headers={
            "Authorization": "Bearer " + create_test_token(scopes=None)  # type: ignore
        },
    ) as ac:
        response = await ac.get("/api/protected/scope")
    assert response.status_code == 401
    assert response.json() == {
        "detail": {"error": "invalid_token", "message": "Token contains invalid formatted scopes"}
    }
    assert response.headers["WWW-Authenticate"] == "Bearer"


async def test_expired_token(fastapi_app, mock_openid_and_keys):
    """Test that an expired token is rejected."""
    async with AsyncClient(
        transport=ASGITransport(app=app),
        base_url="http://test",
        headers={"Authorization": "Bearer " + create_test_token(expired=True)},
    ) as ac:
        response = await ac.get("/api/protected/scope")
    assert response.status_code == 401
    assert response.json() == {"detail": {"error": "invalid_token", "message": "Token signature has expired"}}
    assert response.headers["WWW-Authenticate"] == "Bearer"


async def test_token_signed_with_evil_key(fastapi_app, mock_openid_and_keys):
    """Test that a token signed with an 'evil' key is rejected."""
    async with AsyncClient(
        transport=ASGITransport(app=app),
        base_url="http://test",
        headers={"Authorization": "Bearer " + create_test_token(role="admin", evil=True)},
    ) as ac:
        response = await ac.get("/api/protected/admin")
    assert response.status_code == 401
    assert response.json() == {"detail": {"error": "invalid_token", "message": "Unable to validate token"}}
    assert response.headers["WWW-Authenticate"] == "Bearer"


async def test_malformed_token(fastapi_app, mock_openid_and_keys):
    """Test that a malformed token is rejected."""
    async with AsyncClient(
        transport=ASGITransport(app=app),
        base_url="http://test",
        headers={"Authorization": "Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9"},
    ) as ac:
        response = await ac.get("/api/protected/admin")
    assert response.status_code == 401
    assert response.json() == {"detail": {"error": "invalid_token", "message": "Invalid token format"}}
    assert response.headers["WWW-Authenticate"] == "Bearer"


async def test_none_token(fastapi_app, mock_openid_and_keys, mocker):
    """Test that when no token is available in the request, it is rejected."""
    mocker.patch.object(ZitadelAuth, "_extract_access_token", return_value=None)
    async with AsyncClient(
        transport=ASGITransport(app=app),
        base_url="http://test",
    ) as ac:
        response = await ac.get("/api/protected/admin")
        assert response.status_code == 401
        assert response.json() == {"detail": {"error": "invalid_token", "message": "No access token provided"}}
        assert response.headers["WWW-Authenticate"] == "Bearer"


async def test_missing_authorization_header(fastapi_app, mock_openid_and_keys):
    """Test that when Authorization header is completely missing, request is rejected with 401."""
    async with AsyncClient(
        transport=ASGITransport(app=app),
        base_url="http://test",
    ) as ac:
        response = await ac.get("/api/protected/admin")

        # FastAPI's OAuth2AuthorizationCodeBearer handles missing headers
        assert response.status_code == 401
        assert response.json() == {"detail": "Not authenticated"}
        assert response.headers["WWW-Authenticate"] == "Bearer"


async def test_token_not_bearer(fastapi_app, mock_openid_and_keys, mocker):
    """Test that when the token is not a Bearer token, it is rejected."""
    async with AsyncClient(
        transport=ASGITransport(app=app),
        base_url="http://test",
        headers={"Authorization": create_test_token(role="admin")},
    ) as ac:
        response = await ac.get("/api/protected/admin")

        # actually raised in fastapi.oauth2.OAuth2AuthorizationCodeBearer
        assert response.json() == {"detail": "Not authenticated"}
        assert response.status_code == 401
        assert response.headers["WWW-Authenticate"] == "Bearer"


async def test_token_extraction_raises(fastapi_app, mock_openid_and_keys, mocker):
    """Test that an exception during token extraction is handled."""
    mocker.patch.object(ZitadelAuth, "_extract_access_token", side_effect=ValueError("oops"))
    async with AsyncClient(
        transport=ASGITransport(app=app),
        base_url="http://test",
        headers={"Authorization": "Bearer " + create_test_token(role="admin")},
    ) as ac:
        response = await ac.get("/api/protected/admin")
        assert response.status_code == 400
        assert response.json() == {
            "detail": {"error": "invalid_request", "message": "Unable to extract token from request"}
        }


async def test_header_invalid_alg(fastapi_app, mock_openid_and_keys):
    """Test that a token header with an invalid algorithm is rejected."""
    async with AsyncClient(
        transport=ASGITransport(app=app),
        base_url="http://test",
        headers={"Authorization": "Bearer " + create_test_token(alg="RS512")},
    ) as ac:
        response = await ac.get("/api/protected/admin")
        assert response.status_code == 401
        assert response.json() == {"detail": {"error": "invalid_token", "message": "Invalid token header"}}
        assert response.headers["WWW-Authenticate"] == "Bearer"


async def test_header_invalid_typ(fastapi_app, mock_openid_and_keys):
    """Test that a token header with an invalid type is rejected."""
    async with AsyncClient(
        transport=ASGITransport(app=app),
        base_url="http://test",
        headers={"Authorization": "Bearer " + create_test_token(typ="JWS")},
    ) as ac:
        response = await ac.get("/api/protected/admin")
        assert response.status_code == 401
        assert response.json() == {"detail": {"error": "invalid_token", "message": "Invalid token header"}}
        assert response.headers["WWW-Authenticate"] == "Bearer"


async def test_exception_handled(fastapi_app, mock_openid_and_keys, mocker):
    """Test that an exception during token verification is handled."""
    mocker.patch.object(TokenValidator, "verify", side_effect=ValueError("oops"))
    async with AsyncClient(
        transport=ASGITransport(app=app),
        base_url="http://test",
        headers={"Authorization": "Bearer " + create_test_token(role="admin")},
    ) as ac:
        response = await ac.get("/api/protected/admin")
        assert response.status_code == 401
        assert response.json() == {"detail": {"error": "invalid_token", "message": "Unable to process token"}}
        assert response.headers["WWW-Authenticate"] == "Bearer"


async def test_refresh_config_on_unknown_key_id(fastapi_app, mock_openid_empty_then_ok, mocker):
    """Test that the OpenID configuration is refreshed if the key ID is initially not found."""
    sleep_mock = mocker.patch("fastapi_zitadel_auth.openid_config.OpenIdConfig._sleep")

    async with AsyncClient(
        transport=ASGITransport(app=app),
        base_url="http://test",
        headers={"Authorization": "Bearer " + create_test_token(role="admin", kid="test-key-2")},
    ) as ac:
        response = await ac.get("/api/protected/admin")
        assert response.status_code == 200
        assert sleep_mock.call_count == 1
        sleep_mock.assert_called_once()



"""
Test suite to test the public endpoints.
"""

import pytest


def test_public_endpoint(public_client):
    """Test that the public endpoint works"""
    response = public_client.get("/api/public")
    assert response.status_code == 200, response.text
    assert response.json() == {"message": "Hello world!"}


@pytest.mark.parametrize(
    "path",
    [
        "/api/protected/admin",
        "/api/protected/scope",
    ],
)
def test_no_token(public_client, path):
    """Test that protected endpoints return 401 when no token is provided"""
    response = public_client.get(path)
    assert response.status_code == 401, response.text
    assert response.json() == {"detail": "Not authenticated"}
    assert response.headers["WWW-Authenticate"] == "Bearer"


@pytest.mark.parametrize(
    "path",
    [
        "/api/protected/admin",
        "/api/protected/scope",
    ],
)
def test_incorrect_token(public_client, path):
    """Test that protected endpoints return 401 when an incorrect token is provided"""
    response = public_client.get(path, headers={"Authorization": "Non-existent testtoken"})
    assert response.status_code == 401, response.text
    assert response.json() == {"detail": "Not authenticated"}
    assert response.headers["WWW-Authenticate"] == "Bearer"


@pytest.mark.parametrize(
    "path",
    [
        "/api/protected/admin",
        "/api/protected/scope",
    ],
)
def test_token_empty(public_client, path):
    """Test that protected endpoints return 401 when an empty token is provided"""
    response = public_client.get(path, headers={"Authorization": "Bearer "})
    assert response.status_code == 401, response.text
    assert response.json() == {"detail": {"error": "invalid_token", "message": "Invalid token format"}}
    assert response.headers["WWW-Authenticate"] == "Bearer"



"""
Test the provider configuration (Zitadel) fetching.
"""

import asyncio
import logging
from datetime import datetime, timedelta

import httpx
import pytest
import respx
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from httpx import AsyncClient, ASGITransport, HTTPError

from demo_project.dependencies import zitadel_auth
from demo_project.main import app
from fastapi_zitadel_auth.exceptions import UnauthorizedException
from tests.conftest import openid_configuration
from tests.utils import create_test_token, openid_config_url, keys_url, create_openid_keys

log = logging.getLogger(__name__)


async def test_successful_config_load(mock_openid_and_keys):
    """Test that OpenIdConfig loads config and keys correctly"""

    test_start = datetime.now()
    assert len(zitadel_auth.openid_config.signing_keys) == 0
    assert zitadel_auth.openid_config.last_refresh_timestamp is None

    await zitadel_auth.openid_config.load_config()

    assert len(zitadel_auth.openid_config.signing_keys) == 1
    assert isinstance(zitadel_auth.openid_config.signing_keys, dict)
    assert isinstance(zitadel_auth.openid_config.signing_keys["test-key-1"], RSAPublicKey)

    assert zitadel_auth.openid_config.last_refresh_timestamp is not None
    assert isinstance(zitadel_auth.openid_config.last_refresh_timestamp, datetime)
    assert test_start <= zitadel_auth.openid_config.last_refresh_timestamp <= datetime.now()

    assert zitadel_auth.openid_config.issuer_url
    assert zitadel_auth.openid_config.config_url
    assert zitadel_auth.openid_config.authorization_url
    assert zitadel_auth.openid_config.token_url
    assert zitadel_auth.openid_config.jwks_uri
    assert isinstance(zitadel_auth.openid_config.cache_ttl_seconds, int)
    assert zitadel_auth.openid_config.cache_ttl_seconds > 0


async def test_caching_behavior():
    """Test that caching works and prevents unnecessary refreshes"""
    with respx.mock(assert_all_called=True) as mock:
        mock.get(openid_config_url()).mock(return_value=httpx.Response(200, json=openid_configuration()))
        mock.get(keys_url()).mock(return_value=httpx.Response(200, json=create_openid_keys()))

        # First load
        await zitadel_auth.openid_config.load_config()
        first_timestamp = zitadel_auth.openid_config.last_refresh_timestamp

        # Second immediate load should use cache
        await zitadel_auth.openid_config.load_config()
        assert zitadel_auth.openid_config.last_refresh_timestamp == first_timestamp
        assert mock.calls.call_count == 2  # Only called once for initial load


async def test_cache_expiration():
    """Test that cache refreshes after expiration"""
    with respx.mock(assert_all_called=True) as mock:
        mock.get(openid_config_url()).mock(return_value=httpx.Response(200, json=openid_configuration()))
        mock.get(keys_url()).mock(return_value=httpx.Response(200, json=create_openid_keys()))

        # First load
        await zitadel_auth.openid_config.load_config()

        # Manually expire cache
        zitadel_auth.openid_config.last_refresh_timestamp = datetime.now() - timedelta(
            seconds=zitadel_auth.openid_config.cache_ttl_seconds + 1
        )

        # Should trigger refresh
        await zitadel_auth.openid_config.load_config()
        assert mock.calls.call_count == 4  # Two calls for each load


async def test_concurrent_refresh_requests():
    """Test that concurrent refreshes are handled correctly"""
    with respx.mock(assert_all_called=True) as mock:

        async def slow_config_response(*args, **kwargs):
            await asyncio.sleep(0.2)
            return httpx.Response(200, json=openid_configuration())

        async def slow_keys_response(*args, **kwargs):
            await asyncio.sleep(0.2)
            return httpx.Response(200, json=create_openid_keys())

        config_route = mock.get(openid_config_url()).mock(side_effect=slow_config_response)
        keys_route = mock.get(keys_url()).mock(side_effect=slow_keys_response)

        zitadel_auth.openid_config.reset_cache()

        tasks = [zitadel_auth.openid_config.load_config() for _ in range(5)]
        await asyncio.gather(*tasks)

        assert len(config_route.calls) == 1, "Config endpoint called multiple times"
        assert len(keys_route.calls) == 1, "Keys endpoint called multiple times"
        assert len(zitadel_auth.openid_config.signing_keys) == 1


async def test_provider_connection_failure_openid(fastapi_app):
    """Test that the app handles connection failures to the provider on OpenID config fetch"""
    with respx.mock(assert_all_called=True) as mock:
        mock.get(openid_config_url()).mock(side_effect=HTTPError("Connection failed"))

        async with AsyncClient(
            transport=ASGITransport(app=app),
            base_url="http://test",
            headers={"Authorization": "Bearer " + create_test_token(role="admin")},
        ) as ac:
            response = await ac.get("/api/protected/admin")

            assert response.status_code == 401
            assert response.json() == {
                "detail": {
                    "error": "invalid_token",
                    "message": "Unable to refresh configuration from identity provider",
                }
            }
            assert response.headers["WWW-Authenticate"] == "Bearer"

        assert mock.calls.call_count == 1
        assert zitadel_auth.openid_config.last_refresh_timestamp is None
        assert len(zitadel_auth.openid_config.signing_keys) == 0


async def test_provider_connection_failure_jwks(fastapi_app):
    """Test that the app handles connection failures to the provider on JWKS fetch"""
    with respx.mock(assert_all_called=True) as mock:
        mock.get(openid_config_url()).mock(return_value=httpx.Response(200, json=openid_configuration()))
        mock.get(keys_url()).mock(side_effect=HTTPError("Connection failed"))

        async with AsyncClient(
            transport=ASGITransport(app=app),
            base_url="http://test",
            headers={"Authorization": "Bearer " + create_test_token(role="admin")},
        ) as ac:
            response = await ac.get("/api/protected/admin")

            assert response.status_code == 401
            assert response.json() == {
                "detail": {
                    "error": "invalid_token",
                    "message": "Unable to refresh configuration from identity provider",
                }
            }
            assert response.headers["WWW-Authenticate"] == "Bearer"

        assert mock.calls.call_count == 2
        assert zitadel_auth.openid_config.last_refresh_timestamp is None
        assert len(zitadel_auth.openid_config.signing_keys) == 0


async def test_invalid_config_response():
    """Test handling of invalid config response"""
    with respx.mock(assert_all_called=True) as mock:
        mock.get(openid_config_url()).mock(return_value=httpx.Response(200, json={"invalid": "config"}))

        with pytest.raises(UnauthorizedException):
            await zitadel_auth.openid_config.load_config()


async def test_invalid_jwks_response():
    """Test handling of invalid JWKS response"""
    with respx.mock(assert_all_called=True) as mock:
        mock.get(openid_config_url()).mock(return_value=httpx.Response(200, json=openid_configuration()))
        mock.get(keys_url()).mock(return_value=httpx.Response(200, json={"keys": [{"invalid": "key"}]}))

        async with AsyncClient(
            transport=ASGITransport(app=app),
            base_url="http://test",
            headers={"Authorization": "Bearer " + create_test_token(role="admin")},
        ) as ac:
            response = await ac.get("/api/protected/admin")

            assert response.status_code == 401



"""
Authentication module for Zitadel OAuth2
"""

import logging
from typing import TYPE_CHECKING, Type

from fastapi.exceptions import HTTPException
from fastapi.security import OAuth2AuthorizationCodeBearer, SecurityScopes
from fastapi.security.base import SecurityBase
from jwt import (
    ExpiredSignatureError,
    ImmatureSignatureError,
    InvalidAudienceError,
    InvalidIssuedAtError,
    InvalidIssuerError,
    InvalidTokenError,
    MissingRequiredClaimError,
)
from pydantic import HttpUrl
from starlette.requests import Request

from .exceptions import UnauthorizedException, InvalidRequestException, ForbiddenException
from .user import (
    ClaimsT,
    DefaultZitadelClaims,
    DefaultZitadelUser,
    UserT,
    JwtClaims,
    BaseZitadelUser,
)
from .openid_config import OpenIdConfig
from .token import TokenValidator

if TYPE_CHECKING:  # pragma: no cover
    from jwt.algorithms import AllowedPublicKeys  # noqa: F401

log = logging.getLogger("fastapi_zitadel_auth")


class ZitadelAuth(SecurityBase):
    """
    Zitadel OAuth2 authentication using bearer token
    """

    def __init__(
        self,
        issuer_url: HttpUrl | str,
        project_id: str,
        app_client_id: str,
        allowed_scopes: dict[str, str],
        token_leeway: float = 0,
        cache_ttl_seconds: int = 600,
        claims_model: Type[ClaimsT] = DefaultZitadelClaims,  # type: ignore
        user_model: Type[UserT] = DefaultZitadelUser,  # type: ignore
        scheme_name: str = "ZitadelAuthorizationCodeBearer",
        description: str = "Zitadel OAuth2 authentication using bearer token",
    ) -> None:
        """
        Initialize the ZitadelAuth object

        :param issuer_url: HttpUrl | str
            The Zitadel issuer URL

        :param project_id: str
            The Zitadel project ID

        :param app_client_id: str
            The Zitadel application client ID

        :param allowed_scopes: dict[str, str]
            The allowed scopes for the application. Key is the scope name and value is the description.
            Example:
                {
                    "read": "Read access"
                }

        :param token_leeway: float
            The tolerance time in seconds for token validation

        :param cache_ttl_seconds: int
            The time in seconds to cache the OpenID configuration

        :param claims_model: Type[ClaimsT]
            The claims model to use, e.g. DefaultZitadelClaims. See user.py

        :param user_model:  Type[UserT]
            The user model to use, e.g. DefaultZitadelUser. See user.py

        :param scheme_name: str
            The name of the security scheme for OpenAPI documentation.
            Default: "ZitadelAuthorizationCodeBearer"

        :param description: str
            The description of the security scheme for OpenAPI documentation.
            Default: "Zitadel OAuth2 authentication using bearer token"
        """

        self.client_id = app_client_id
        self.project_id = project_id
        self.issuer_url = str(issuer_url).rstrip("/")
        self.token_leeway = token_leeway

        if not issubclass(claims_model, JwtClaims):
            raise ValueError("claims_model must be a subclass of JwtClaims")

        if not issubclass(user_model, BaseZitadelUser):
            raise ValueError("user_model must be a subclass of BaseZitadelUser")

        if not isinstance(scheme_name, str) or not scheme_name.strip():
            raise ValueError("scheme_name must be a non-empty string")

        if not isinstance(description, str) or not description.strip():
            raise ValueError("description must be a non-empty string")

        self.claims_model = claims_model
        self.user_model = user_model

        self.openid_config = OpenIdConfig(
            issuer_url=self.issuer_url,
            config_url=f"{self.issuer_url}/.well-known/openid-configuration",
            authorization_url=f"{self.issuer_url}/oauth/v2/authorize",
            token_url=f"{self.issuer_url}/oauth/v2/token",
            jwks_uri=f"{self.issuer_url}/oauth/v2/keys",
            cache_ttl_seconds=cache_ttl_seconds,
        )

        self.oauth = OAuth2AuthorizationCodeBearer(
            authorizationUrl=self.openid_config.authorization_url,
            tokenUrl=self.openid_config.token_url,
            scopes=allowed_scopes,
            scheme_name=scheme_name,
            description=description,
        )

        self.token_validator = TokenValidator()
        self.model = self.oauth.model
        self.scheme_name = self.oauth.scheme_name

    async def __call__(self, request: Request, security_scopes: SecurityScopes) -> UserT | None:
        """
        Extend the SecurityBase.__call__ method to validate the Zitadel OAuth2 token.
        see also FastAPI -> "Advanced Dependency".
        """
        try:
            access_token = await self._extract_access_token(request)
            if access_token is None:
                raise UnauthorizedException("No access token provided")

            unverified_header, unverified_claims = self.token_validator.parse_unverified_token(access_token)
            self.token_validator.validate_header(unverified_header)
            self.token_validator.validate_scopes(unverified_claims, security_scopes.scopes)

            await self.openid_config.load_config()
            signing_key = await self.openid_config.get_key(unverified_header["kid"])

            try:
                verified_claims = self.token_validator.verify(
                    token=access_token,
                    key=signing_key,
                    audiences=[self.client_id, self.project_id],
                    issuer=self.openid_config.issuer_url,
                    token_leeway=self.token_leeway,
                )
            except (
                InvalidAudienceError,
                InvalidIssuerError,
                InvalidIssuedAtError,
                ImmatureSignatureError,
                MissingRequiredClaimError,
            ) as error:
                log.info(f"Token contains invalid claims: {error}")
                raise UnauthorizedException("Token contains invalid claims") from error

            except ExpiredSignatureError as error:
                log.info(f"Token signature has expired. {error}")
                raise UnauthorizedException("Token signature has expired") from error

            except InvalidTokenError as error:
                log.warning(f"Invalid token. Error: {error}", exc_info=True)
                raise UnauthorizedException("Unable to validate token") from error

            except Exception as error:
                # Extra failsafe in case of a bug in PyJWT
                log.exception(f"Unable to process jwt token. Uncaught error: {error}")
                raise UnauthorizedException("Unable to process token") from error

            else:
                user: UserT = self.user_model(  # type: ignore
                    claims=self.claims_model.model_validate(verified_claims),
                    access_token=access_token,
                )
                # Add the user to the request state
                request.state.user = user
                return user

        except (UnauthorizedException, InvalidRequestException, ForbiddenException, HTTPException):
            raise

        except Exception as error:
            # Failsafe in case of error in OAuth2AuthorizationCodeBearer.__call__
            log.warning(f"Unable to extract token from request. Error: {error}")
            raise InvalidRequestException("Unable to extract token from request") from error

    async def _extract_access_token(self, request: Request) -> str | None:
        """
        Extract the access token from the request
        """
        return await self.oauth(request=request)



"""
Tests for OpenIdConfig class
"""

import pytest
from datetime import datetime, timedelta
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey

from fastapi_zitadel_auth.openid_config import OpenIdConfig
from tests.utils import valid_key, ZITADEL_ISSUER, openid_config_url, keys_url


@pytest.fixture
def mock_openid_config():
    """Fixture providing mock OpenID configuration data."""
    return {
        "issuer": ZITADEL_ISSUER,
        "authorization_endpoint": f"{ZITADEL_ISSUER}/oauth/v2/authorize",
        "token_endpoint": f"{ZITADEL_ISSUER}/oauth/v2/token",
        "jwks_uri": f"{ZITADEL_ISSUER}/oauth/v2/keys",
    }


@pytest.fixture
def mock_jwks():
    """Fixture providing mock JWKS data"""
    return {
        "keys": [
            {
                "kty": "RSA",
                "kid": "305785621098529823",
                "use": "sig",
                "alg": "RS256",
                "n": "sample_n",
                "e": "AQAB",
            },
            {
                "kty": "RSA",
                "kid": "305785621098529824",
                "use": "sig",
                "alg": "RS256",
                "n": "sample_n_2",
                "e": "AQAB",
            },
            {"kty": "RSA", "kid": "invalid_key", "use": "enc", "alg": "RS512"},
        ]
    }


@pytest.mark.asyncio
class TestOpenIdConfig:
    """Test suite for OpenIdConfig class"""

    async def test_successful_config_load(self, respx_mock, mock_openid_config, mock_jwks):
        """Test that OpenIdConfig loads config and keys correctly"""
        config_url = openid_config_url()
        config = OpenIdConfig(
            issuer_url="",
            config_url=config_url,
            authorization_url="",
            token_url="",
            jwks_uri=keys_url(),
        )

        respx_mock.get(config_url).respond(json=mock_openid_config)
        respx_mock.get(mock_openid_config["jwks_uri"]).respond(json=mock_jwks)

        await config.load_config()

        assert config.issuer_url == mock_openid_config["issuer"]
        assert config.authorization_url == mock_openid_config["authorization_endpoint"]
        assert config.token_url == mock_openid_config["token_endpoint"]
        assert config.jwks_uri == mock_openid_config["jwks_uri"]
        assert isinstance(config.cache_ttl_seconds, int)
        assert len(config.signing_keys) == 2
        assert all(isinstance(key, RSAPublicKey) for key in config.signing_keys.values())

    @pytest.mark.parametrize("cache_ttl_seconds", [1, 30, 100, 300, 600, 3600])
    async def test_caching_behavior(self, cache_ttl_seconds, respx_mock, mock_openid_config, mock_jwks, freezer):
        """Test that config is cached and only refreshed after expiry"""
        config_url = openid_config_url()
        config = OpenIdConfig(
            issuer_url="",
            config_url=config_url,
            authorization_url="",
            token_url="",
            jwks_uri=keys_url(),
            cache_ttl_seconds=cache_ttl_seconds,
        )

        initial_config_request = respx_mock.get(config_url).respond(json=mock_openid_config)
        initial_jwks_request = respx_mock.get(mock_openid_config["jwks_uri"]).respond(json=mock_jwks)

        # Load config and keys at a fixed time
        start_datetime = datetime(2025, 2, 5, 18, 0, 0)
        freezer.move_to(start_datetime)
        await config.load_config()
        initial_refresh = config.last_refresh_timestamp
        assert initial_config_request.call_count == 1
        assert initial_jwks_request.call_count == 1

        # Move time forward by a second
        freezer.move_to(start_datetime + timedelta(seconds=1))
        await config.load_config()  # Should use cached config
        assert config.last_refresh_timestamp == initial_refresh  # Timestamp shouldn't change for cache hit
        assert initial_config_request.call_count == 1  # Should not have changed
        assert initial_jwks_request.call_count == 1  # Should not have changed

        # Move time forward past cache expiration
        freezer.move_to(start_datetime + timedelta(seconds=cache_ttl_seconds + 1))
        await config.load_config()  # Should refresh
        assert config.last_refresh_timestamp > initial_refresh
        assert initial_config_request.call_count == 2  # Should have refreshed
        assert initial_jwks_request.call_count == 2  # Should have refreshed

    async def test_key_filtering(self, respx_mock, mock_openid_config):
        """Test that invalid keys are filtered out"""
        config_url = openid_config_url()
        config = OpenIdConfig(
            issuer_url="",
            config_url=config_url,
            authorization_url="",
            token_url="",
            jwks_uri=keys_url(),
        )

        invalid_jwks = {
            "keys": [
                {
                    "kty": "EC",
                    "kid": "1",
                    "use": "sig",
                    "alg": "ES256",
                },  # Wrong key type
                {"kty": "RSA", "use": "sig", "alg": "RS256"},  # Missing kid
                {"kty": "RSA", "kid": "3", "use": "enc", "alg": "RS256"},  # Wrong use
                {"kty": "RSA", "kid": "4", "use": "sig", "alg": "RS512"},  # Wrong alg
            ]
        }

        respx_mock.get(config_url).respond(json=mock_openid_config)
        respx_mock.get(mock_openid_config["jwks_uri"]).respond(json=invalid_jwks)

        await config.load_config()
        assert len(config.signing_keys) == 0

    @pytest.mark.parametrize(
        "last_refresh_timestamp, signing_key, expected",
        [
            (None, {}, True),  # No config -> refresh
            (datetime.now(), {}, True),  # No keys -> refresh
            (
                datetime.now(),
                valid_key.public_key(),
                False,
            ),  # Fresh config and keys -> no refresh
            (
                datetime.now() - timedelta(hours=2),
                valid_key.public_key(),
                True,
            ),  # Old config -> refresh
            (None, valid_key.public_key(), True),  # No config, but keys -> refresh
        ],
    )
    async def test_needs_refresh(self, last_refresh_timestamp, signing_key, expected):
        """Test that _needs_refresh method works as expected based on last_refresh_timestamp and signing_keys"""
        config_url = openid_config_url()
        config = OpenIdConfig(
            issuer_url="",
            config_url=config_url,
            authorization_url="",
            token_url="",
            jwks_uri="",
            last_refresh_timestamp=last_refresh_timestamp,
            signing_keys={"kid": signing_key} if signing_key else {},
        )
        assert config._needs_refresh() == expected



"""
Test the OpenAPI schema.
"""

import openapi_spec_validator
from packaging import version

import fastapi
from fastapi_zitadel_auth import __version__

from demo_project.main import app
from tests.utils import ZITADEL_ISSUER

expected_scheme_name = "ZitadelAuth"
expected_description = "OAuth2 Authorization Code Flow with PKCE for Zitadel authentication"

# FastAPI 0.123.8+ changed how scopes are represented in OpenAPI schema
# https://github.com/fastapi/fastapi/pull/14455
# https://github.com/fastapi/fastapi/releases/tag/0.123.8
FASTAPI_VERSION = version.parse(fastapi.__version__)
SCOPES_IN_SECURITY = ["scope1"] if FASTAPI_VERSION >= version.parse("0.123.8") else []

openapi_schema = {
    "openapi": "3.1.0",
    "info": {"title": "fastapi-zitadel-auth demo", "version": __version__},
    "paths": {
        "/api/public": {
            "get": {
                "summary": "Public endpoint",
                "description": "Public endpoint",
                "operationId": "public_api_public_get",
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    }
                },
            }
        },
        "/api/protected/admin": {
            "get": {
                "summary": "Protected endpoint, requires admin role",
                "description": "Protected endpoint",
                "operationId": "protected_for_admin_api_protected_admin_get",
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    }
                },
                "security": [{expected_scheme_name: []}],
            }
        },
        "/api/protected/scope": {
            "get": {
                "summary": "Protected endpoint, requires a specific scope",
                "description": "Protected endpoint, requires a specific scope",
                "operationId": "protected_by_scope_api_protected_scope_get",
                "responses": {
                    "200": {
                        "description": "Successful Response",
                        "content": {"application/json": {"schema": {}}},
                    }
                },
                "security": [{expected_scheme_name: SCOPES_IN_SECURITY}],
            }
        },
    },
    "components": {
        "securitySchemes": {
            expected_scheme_name: {
                "type": "oauth2",
                "description": expected_description,
                "flows": {
                    "authorizationCode": {
                        "scopes": {
                            "openid": "OpenID Connect",
                            "email": "Email",
                            "profile": "Profile",
                            "urn:zitadel:iam:org:project:id:zitadel:aud": "Audience",
                            "urn:zitadel:iam:org:projects:roles": "Projects roles",
                        },
                        "authorizationUrl": f"{ZITADEL_ISSUER}/oauth/v2/authorize",
                        "tokenUrl": f"{ZITADEL_ISSUER}/oauth/v2/token",
                    }
                },
            }
        }
    },
}


def test_openapi_schema(public_client):
    """Test the OpenAPI schema matches to the expected schema"""
    response = public_client.get("/openapi.json")
    assert response.status_code == 200, response.text
    print(response.json())
    assert response.json() == openapi_schema


def test_validate_openapi_spec(public_client):
    """Validate the OpenAPI spec"""
    response = public_client.get("/openapi.json")
    assert response.status_code == 200, response.text
    openapi_spec_validator.validate(response.json())


def test_custom_scheme_name_and_description():
    """Test that custom scheme_name and description from demo app appear in OpenAPI schema"""

    openapi_schema_from_app = app.openapi()
    assert expected_scheme_name in openapi_schema_from_app["components"]["securitySchemes"]
    assert (
        openapi_schema_from_app["components"]["securitySchemes"][expected_scheme_name]["description"]
        == expected_description
    )



"""
Settings for the demo project
"""

from functools import lru_cache
from typing import Literal

from pydantic import HttpUrl
from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    ZITADEL_HOST: HttpUrl
    ZITADEL_PROJECT_ID: str
    OAUTH_CLIENT_ID: str
    LOG_LEVEL: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = "DEBUG"

    model_config = SettingsConfigDict(env_file="demo_project/.env", env_file_encoding="utf-8")


@lru_cache
def get_settings() -> Settings:
    """Singleton function to load settings."""
    return Settings()



import logging
from typing import Any

import jwt
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey

from fastapi_zitadel_auth.exceptions import UnauthorizedException, ForbiddenException

log = logging.getLogger("fastapi_zitadel_auth")


class TokenValidator:
    """Handles JWT token validation and parsing"""

    @staticmethod
    def validate_scopes(claims: dict[str, Any], required_scopes: list[str] | None) -> bool:
        """
        Validates that the token has the required scopes
        """
        if required_scopes is None:
            return True

        # Check if the token has the scope field and it is a string
        token_scope_str = claims.get("scope", "")
        if not isinstance(token_scope_str, str):
            log.warning("Invalid scope format: %s", token_scope_str)
            raise UnauthorizedException("Token contains invalid formatted scopes")
        token_scopes = token_scope_str.split()

        # Check if all required scopes are present
        for required_scope in required_scopes:
            if required_scope not in token_scopes:
                log.debug(f"Missing required scope: {required_scope}. Available scopes: {token_scopes}")
                raise ForbiddenException(f"Missing required scope: {required_scope}")
        return True

    @staticmethod
    def parse_unverified_token(
        access_token: str,
    ) -> tuple[dict[str, Any], dict[str, Any]]:
        """Parse header and claims without verification"""
        try:
            header = dict(jwt.get_unverified_header(access_token))
            claims = dict(jwt.decode(access_token, options={"verify_signature": False}))
            return header, claims
        except Exception as e:
            log.warning(
                "Malformed token received. %s. Error: %s",
                access_token,
                e,
                exc_info=True,
            )
            raise UnauthorizedException("Invalid token format") from e

    @staticmethod
    def validate_header(header: dict[str, Any]) -> None:
        """Validate token header"""
        if header.get("alg") != "RS256" or header.get("typ") != "JWT":
            raise UnauthorizedException("Invalid token header")

    @staticmethod
    def verify(
        token: str,
        key: RSAPublicKey,
        audiences: list[str],
        issuer: str,
        token_leeway: float = 0,
    ) -> dict[str, Any]:
        """Verify token signature and claims with provided key"""
        options = {
            "verify_signature": True,
            "verify_exp": True,
            "verify_nbf": True,
            "verify_iat": True,
            "verify_aud": True,
            "verify_iss": True,
            "verify_sub": True,
            "verify_jti": True,
            "require": ["exp", "iat", "aud", "iss", "sub"],
        }
        return jwt.decode(
            token,
            key=key,
            algorithms=["RS256"],
            audience=audiences,
            issuer=issuer,
            leeway=token_leeway,
            options=options,
        )



import json
import time
import os
from pathlib import Path
import jwt as pyjwt
from locust import HttpUser, task, events, between
from locust.runners import MasterRunner
import logging


ZITADEL_HOST = os.getenv("ZITADEL_HOST", "https://zitadel-instance.example.com")
ZITADEL_PROJECT_ID = os.getenv("ZITADEL_PROJECT_ID", "project-id")
ZITADEL_CLIENT_ID = os.getenv("OAUTH_CLIENT_ID", "client-id")
LOG_LEVEL = os.getenv("LOG_LEVEL", "INFO")
SERVICE_USER_PRIVATE_KEY_FILE = os.getenv("SERVICE_USER_KEY_FILE", "service_user.json")
ROLE = os.getenv("ZITADEL_ROLE", "admin")
API_HOST = os.getenv("API_HOST", "http://localhost:8001")

# Set up logging
logging.basicConfig(level=getattr(logging, LOG_LEVEL))
logger = logging.getLogger(__name__)

# Log all environment variables being used
logger.info(f"ZITADEL_HOST: {ZITADEL_HOST}")
logger.info(f"ZITADEL_PROJECT_ID: {ZITADEL_PROJECT_ID}")
logger.info(f"OAUTH_CLIENT_ID: {ZITADEL_CLIENT_ID}")
logger.info(f"LOG_LEVEL: {LOG_LEVEL}")
logger.info(f"SERVICE_USER_KEY_FILE: {SERVICE_USER_PRIVATE_KEY_FILE}")
logger.info(f"ZITADEL_ROLE: {ROLE}")
logger.info(f"API_HOST: {API_HOST}")


# Set up temp file to store token
TOKEN_PATH = Path("/mnt/locust/token.txt")


def load_token() -> str:
    """Load token from file if it exists"""
    try:
        return TOKEN_PATH.read_text().strip()
    except FileNotFoundError:
        return ""


def write_token(token: str) -> None:
    """Write token to file"""
    TOKEN_PATH.write_text(token)


def delete_token() -> None:
    """Delete token file if it exists"""
    try:
        TOKEN_PATH.unlink(missing_ok=True)
    except (PermissionError, FileNotFoundError):
        pass


def get_token_from_zitadel() -> str:
    """Generate a new token without counting as a Locust request"""
    try:
        key_file_path = Path(SERVICE_USER_PRIVATE_KEY_FILE)
        if not key_file_path.exists():
            raise FileNotFoundError(f"Service user key file not found: {SERVICE_USER_PRIVATE_KEY_FILE}")
        with open(key_file_path, "r") as file:
            json_data = json.load(file)

        # Extracting necessary values from the JSON data
        private_key = json_data["key"]
        kid = json_data["keyId"]
        user_id = json_data["userId"]

        # Preparing the JWT header and payload for authentication
        header = {"alg": "RS256", "kid": kid}
        payload = {
            "iss": user_id,
            "sub": user_id,
            "aud": str(ZITADEL_HOST).rstrip("/"),  # Remove trailing slash
            "iat": int(time.time()),
            "exp": int(time.time()) + 3600,  # Token expires in 1 hour
        }

        # Generating JWT token
        jwt_token = pyjwt.encode(payload, private_key, algorithm="RS256", headers=header)

        # Data payload for the OAuth2 token request
        data = {
            "grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer",
            "scope": " ".join(
                [
                    "openid",
                    f"urn:zitadel:iam:org:project:role:{ROLE}",
                    f"urn:zitadel:iam:org:project:id:{ZITADEL_PROJECT_ID}:aud",
                ]
            ),
            "assertion": jwt_token,
        }

        # Use requests directly instead of self.client to avoid counting this as a Locust request
        import requests

        oauth_url = f"{ZITADEL_HOST}/oauth/v2/token"
        response = requests.post(url=oauth_url, data=data)
        response.raise_for_status()

        return response.json()["access_token"]

    except Exception as e:
        logger.error(f"Authentication error: {str(e)}")
        return ""


@events.test_start.add_listener
def on_test_start(environment, **kwargs):
    """
    Initialize token at test start (only for master)
    """
    if isinstance(environment.runner, MasterRunner):
        try:
            # First check if we already have a token
            token = load_token()
            if not token:
                logger.info("Getting new token from Zitadel")
                token = get_token_from_zitadel()
                if token:
                    write_token(token)
                    logger.info("Token acquired and saved successfully")
                else:
                    logger.error("Failed to get token from Zitadel")
                    environment.runner.quit()
        except Exception as e:
            logger.error(f"Failed during test startup: {e}")
            environment.runner.quit()


@events.test_stop.add_listener
def on_test_stop(environment, **kwargs):
    """
    Clean up token on test end
    """
    if isinstance(environment.runner, MasterRunner):
        delete_token()
        logger.info("Token file deleted")


class ZitadelServiceUser(HttpUser):
    wait_time = between(1, 3)  # Wait between 1 and 3 seconds between tasks
    abstract = True  # This is a base class, don't create users of this type

    def on_start(self):
        """
        Get the access token for all requests
        """
        self.token = load_token()
        if not self.token:
            logger.error("No token available for user")

        self.token_timestamp = time.time()

    def refresh_token_if_needed(self):
        """
        Check if the token needs to be refreshed and refresh it if necessary.
        """
        # Refresh token if older than 55 minutes (tokens last 1 hour)
        if time.time() - self.token_timestamp > 55 * 60:
            logger.info("Token is about to expire, refreshing")
            # For simplicity, we'll just read the token again (assuming master has refreshed it)
            new_token = load_token()
            if new_token and new_token != self.token:
                self.token = new_token
                self.token_timestamp = time.time()


class DemoApiUser(ZitadelServiceUser):
    @task(1)
    def access_public_endpoint(self):
        """
        Test accessing a public endpoint (no authentication required).
        """
        with self.client.get(f"{API_HOST}/api/public", name="Public: Get public data", catch_response=True) as response:
            if response.status_code != 200:
                response.failure(f"Public endpoint failed: {response.status_code} - {response.text}")

    @task(2)
    def access_protected_admin_endpoint(self):
        """
        Test accessing a protected endpoint requiring admin role.
        """
        self.refresh_token_if_needed()

        if not self.token:
            logger.warning("Skipping protected endpoint test - no token available")
            return

        headers = {"Authorization": f"Bearer {self.token}"}
        with self.client.get(
            f"{API_HOST}/api/protected/admin", headers=headers, name="Protected: Admin endpoint", catch_response=True
        ) as response:
            if response.status_code != 200:
                response.failure(f"Admin endpoint failed: {response.status_code} - {response.text}")


@events.quitting.add_listener
def _(environment, **kw):
    """Set the exit code to non-zero if any of the following conditions are met:
    - More than 1% of the requests failed
    - The average response time is longer than 200 ms
    - The 95th percentile for response time is larger than 600 ms
    """
    if environment.stats.total.fail_ratio > 0.01:
        logger.error("Test failed due to failure ratio > 1%")
        environment.process_exit_code = 1
    elif environment.stats.total.avg_response_time > 200:
        logger.error("Test failed due to average response time > 200 ms")
        environment.process_exit_code = 1
    elif environment.stats.total.get_response_time_percentile(0.95) > 600:
        logger.error("Test failed due to 95th percentile response time > 600 ms")
        environment.process_exit_code = 1
    else:
        environment.process_exit_code = 0
